/*
Copyright � 2014 NaturalPoint Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. */

// SimpleVRPNTest is a Windows console app that uses VRPN
// to connect to Motive and record the VRPN streaming data from
// a rigid body and an analog device, along with timing information,
// to test Motive's VRPN server.

//
// Uses VRPN v 7.33.1
//

#include <stdio.h>
#include <conio.h>
#include <chrono>
#include <thread>

// VRPN
#include "vrpn_Configure.h"             // for VRPN_CALLBACK, etc
#include "vrpn_Connection.h"			// for test server
#include "vrpn_Types.h"                 // for vrpn_float64
#include "vrpn_Tracker.h"               // for vrpn_TRACKERCB, etc
#include "vrpn_Button.h"
#include "vrpn_Analog.h"

// high resolution timing helper
#include "HiTimer.h"

// constants
const char* kButtonName = "Example1Button";		// "($MotiveButtonDeviceName)Button"
const char* kAnalogName = "Example1Analog";		// "($MotiveAnalogDeviceName)Analog"
const float kTrackerSampleRate = 100.0f;
const float kDeviceSampleRate = 200.0f;
const double kCaptureDuration = 10.0;

const bool kCSVCompatible = true;
const bool kUseLocalVRPNTestServer = false;

// frame drop/dupe detection.  Elapsed time between frame 
// should be ~= the frame period +/- a little fudge room
const double fudgeOver = 1.5;	
const double fudgeUnder = 0.5;
const int kUndefined = -99999;

// output files
FILE* gTrackerFile = nullptr;
FILE* gAnalogFile = nullptr;
FILE* gButtonFile = nullptr;

// timing information
HiTimer timer;
double gInterframeDelta = kUndefined;
int gTrackerDroppedFrames = 0;
int gTrackerDuplicateFrames = 0;
int gTrackerReceivedFrames = 0;
int gButtonDroppedFrames = 0;
int gButtonDuplicateFrames = 0;
int gButtonReceivedFrames = 0;
int gAnalogDroppedFrames = 0;
int gAnalogDuplicateFrames = 0;
int gAnalogReceivedFrames = 0;
double gTrackerPrevTime = kUndefined;
double gButtonPrevTime = kUndefined;
double gAnalogPrevTime = kUndefined;

// test server devices
vrpn_Tracker_NULL* ntkr = nullptr;
vrpn_Tracker_Remote* tkr = nullptr;
vrpn_Analog_Remote* ana = nullptr;
vrpn_Connection* connection = nullptr;

extern const char* vrpn_MAGIC;
static const char* VRPNVersion()
{
	unsigned int i = 0;
	for (i = 0; i < strlen(vrpn_MAGIC); i++)
		if (vrpn_MAGIC[i] >= '0' && vrpn_MAGIC[i] <= '9')
			return &vrpn_MAGIC[i];
	return vrpn_MAGIC;
}

double timestampToMilliseconds(const struct timeval t)
{
	double currentMSecs = ((double)t.tv_sec * 1000.0) + ((double)t.tv_usec * 0.001);
	return currentMSecs;
}

double getTrackerInterFrameDelta()
{
	// first frame
	if (gInterframeDelta == kUndefined)
	{
		gInterframeDelta = 0.0;
		timer.Start();
		return 0.0;
	}

	// time between frames
	timer.Stop();
	gInterframeDelta = timer.Duration();
	timer.Start();

	return gInterframeDelta * 1000.0;
}

static std::string hasDroppedOrDuped(timeval sampleTime, double prevSampleMSecs, double  sampleRate, int& drop, int& dupe, double& outElapsedSampleMSecs)
{
	if (prevSampleMSecs == kUndefined)
	{
		return "";
	}
	
	// record vrpn timestamps for irregularities
	double currentSampleMSecs = timestampToMilliseconds(sampleTime);
	outElapsedSampleMSecs = currentSampleMSecs - prevSampleMSecs;
	double period = (1.0 / sampleRate) * 1000.0;

	std::string hasDroppedOrDuped = "";
	if (outElapsedSampleMSecs > (period * fudgeOver))
	{
		hasDroppedOrDuped = "DROP";
		drop++;
	}
	else if (outElapsedSampleMSecs < (period * fudgeUnder))
	{
		hasDroppedOrDuped = "DUPLICATE";
		dupe++;
	}

	return hasDroppedOrDuped;
}

static void VRPN_CALLBACK handle_tracker(void* userData, const vrpn_TRACKERCB dev)
{
	double sampleRate = kTrackerSampleRate;

	// get local elapsed time
	double interframeMSecs = getTrackerInterFrameDelta();

	// get vrpn elapsed time
	double currentSampleMSecs = timestampToMilliseconds(dev.msg_time);
	double elapsedMSecs = 0.0;
	std::string err = hasDroppedOrDuped(dev.msg_time, gTrackerPrevTime, sampleRate, gTrackerDroppedFrames, gTrackerDuplicateFrames, elapsedMSecs);

	// if vrpn timstamps indicate drop/dupe, report local timing as well
	if(!err.empty())
	{
		double period = (1.0 / sampleRate) * 1000.0;
		if (interframeMSecs > (period * fudgeOver))
		{
			err += " & LOCAL DROP";
		}
		else if (interframeMSecs < (period * fudgeUnder))
		{
			err += " & LOCAL DUPLICATE";
		}
	}

	// record timestamp anomaly
	if (!err.empty())
	{
		// record frame anomaly into file
		if (kCSVCompatible)
		{
			fprintf(gTrackerFile, "%.2f,%.2f,%.2f,%s\n", currentSampleMSecs, elapsedMSecs, interframeMSecs, err.c_str());
		}
		else
		{
			fprintf(gTrackerFile, "\n[%s]\n", err.c_str());
			fprintf(gTrackerFile, "\tVRPN current sample timestamp (secs,usecs) : [%ld][%ld]\n", dev.msg_time.tv_sec, dev.msg_time.tv_usec);
			fprintf(gTrackerFile, "\tVRPN elapsed time from previous sample (msecs) : [%.2f]\n", elapsedMSecs);

		}
	}

	// record frame data
	fprintf(gTrackerFile, "%.2f,%.2f,%.2f,%.3f,%.3f,%.3f,%.3f,%.3f,%.3f,%.3f\n",
		currentSampleMSecs, elapsedMSecs, interframeMSecs, dev.pos[0], dev.pos[1], dev.pos[2], dev.quat[0], dev.quat[1], dev.quat[2], dev.quat[3]);
	
	gTrackerPrevTime = currentSampleMSecs;
	gTrackerReceivedFrames++;

}

static void VRPN_CALLBACK handle_analog(void* userData, const vrpn_ANALOGCB dev)
{
	double elapsedMSecs = 0.0;
	std::string err = hasDroppedOrDuped(dev.msg_time, gAnalogPrevTime, kDeviceSampleRate, gAnalogDroppedFrames, gAnalogDuplicateFrames, elapsedMSecs);

	double currentMSecs = timestampToMilliseconds(dev.msg_time);
	fprintf(gAnalogFile,"%.3f,%.3f,", currentMSecs, elapsedMSecs);
	for (int i = 0; i < dev.num_channel; i++)
	{
		fprintf(gAnalogFile, "%d,%.3f", i, dev.channel[i]);
	}
	fprintf(gAnalogFile, "\n");
	gAnalogPrevTime = currentMSecs;
	gAnalogReceivedFrames++;
}

static void VRPN_CALLBACK handle_button(void* userData, const vrpn_BUTTONCB dev)
{
	// note : VRPN only sends packets when values have changed.
	// In the vase of the button, timing on button off/on state change is discontinuous
	// so lets just measure and report.
	double currentSampleMSecs = timestampToMilliseconds(dev.msg_time);
	double elapsedMSecs = currentSampleMSecs - gButtonPrevTime;
	fprintf(gButtonFile, "%.3f,%.3f,%d,%d\n", currentSampleMSecs, elapsedMSecs, dev.button, dev.state);
	gButtonPrevTime = currentSampleMSecs;
	gButtonReceivedFrames++;
}



static void createOutputFiles(const char* szRootFolder)
{
	char szFile[MAX_PATH];
	sprintf(szFile, "%s\\%s", szRootFolder, "output_tracker.txt");
	gTrackerFile = fopen(szFile, "w");
	if (!gTrackerFile)
	{
		printf("error opening output file %s.  Exiting.", szFile);
	}
	else
	{
		fprintf(gTrackerFile, "VRPN sample timestamp(ms),VRPN sample delta(ms),Local client inter-frame delta(ms), pos x, pos y, pos z, ori x, ori y, ori z, ori w\n");
	}

	sprintf(szFile, "%s\\%s", szRootFolder, "output_analog.txt");
	gAnalogFile = fopen(szFile, "w");
	if (!gAnalogFile)
	{
		printf("error opening output file %s.  Exiting.", szFile);
	}
	else
	{
		fprintf(gAnalogFile, "VRPN sample timestamp(ms),VRPN sample delta(ms),Analog Sensor number, Analog Value\n");
	}

	sprintf(szFile, "%s\\%s", szRootFolder, "output_button.txt");
	gButtonFile = fopen(szFile, "w");
	if (!gButtonFile)
	{
		printf("error opening output file %s.  Exiting.", szFile);
	}
	else
	{
		fprintf(gButtonFile, "VRPN sample timestamp(ms), Elapsed Time (ms), Button number, Button State\n");
	}

}



// SimpleVRPNTest.exe MotiveIPAddress LocalNICIPAddress
// if no addresses specified, 'localhost' is assumed
//Usage Command for Program: exe arg1 (server address) arg2 (local address) arg3 (objectName).
int main(int argc, char* argv[])
{

	printf("Optitrack VRPN Client Test.\n");
	const char* vrpnver = VRPNVersion();
	if (vrpnver)
	{
		printf("VRPN Version : %s\n", VRPNVersion());

	}

	// get Motive address - default to localhost unsless cmd line args present
	//Note: For Command Line Automated Testing, all four values must be defined in the following order: 
	//ServerAddress  LocalAddress  Port  Object Name
	char szServerAddress[MAX_PATH] = "localhost";
	char szLocalAddress[MAX_PATH] = "localhost";
	char szObjectName[MAX_PATH] = "Rigid Body 001";
	char szPortNumber[MAX_PATH] = "3883";
	char vrpnConnectionAddress[MAX_PATH] = "";

	if (argc > 1) {
		if (argc > 1 && argc < 3)
		{
			strcpy_s(szServerAddress, argv[1]);	// Motive server
		}
		if (argc > 2 && argc <4)
		{
			strcpy_s(szLocalAddress, argv[2]);	// Local NIC
		}
		if (argc > 3 && argc < 5)
		{
			strcpy_s(szPortNumber, argv[3]); //Port Number

		}
		if (argc > 4 && argc < 6) {
			strcpy_s(szObjectName, argv[4]); //Rigid Body Object Name
		}
	}
	else {
	//prompting section
	
	char tempServerAddress[MAX_PATH];
	char tempLocalAddress[MAX_PATH];
	char tempPort[MAX_PATH];
	char tempObjectName[MAX_PATH];
	

	printf("\nServer Address (localhost): ");
	fgets(tempServerAddress, MAX_PATH, stdin);
	printf("Local Address (localhost): ");
	fgets(tempLocalAddress, MAX_PATH, stdin);

	printf("Port Number (3883): ");
	fgets(tempPort, MAX_PATH, stdin);
	printf("Object Name (Rigid Body 001): ");
	fgets(tempObjectName, MAX_PATH, stdin);

	//comparison logic that enables for switching default vs user variables. 

	if ((strcmp(szServerAddress, tempServerAddress) != 0) && strcmp(tempServerAddress, "\n") != 0) {
		tempServerAddress[strcspn(tempServerAddress, "\n")] = '\0';
		strcpy_s(szServerAddress, tempServerAddress); //Motive Server
	}
	if ((strcmp(szLocalAddress, tempLocalAddress) != 0) && strcmp(tempLocalAddress, "\n") != 0) {
		tempLocalAddress[strcspn(tempLocalAddress, "\n")] = '\0';
		strcpy_s(szLocalAddress, tempLocalAddress); //Local NIC Address
	}
	if ((strcmp(szPortNumber, tempPort) != 0) && strcmp(tempPort, "\n") != 0) {
		tempPort[strcspn(tempPort, "\n")] = '\0';
		strcpy_s(szPortNumber, tempPort); //Port Number
	}
	if (strcmp(szObjectName, tempObjectName) != 0 && strcmp(tempObjectName, "\n") != 0) {
		tempObjectName[strcspn(tempObjectName, "\n")] = '\0';
		strcpy_s(szObjectName, tempObjectName); //Object Name 
	}
}
	printf("\n\nMotive IP : %s\nLocal IP : %s\nMotive Port Number: %s\nMotive Rigid Body: %s\n\n\n", szServerAddress, szLocalAddress, szPortNumber, szObjectName);
	// create output files for writing received stream into
	char szFolder[MAX_PATH];
	GetCurrentDirectory(MAX_PATH, szFolder);
	createOutputFiles(szFolder);
	printf("Output folder : %s\n", szFolder);

	//create the address that VRPN connection will listen to. 
	sprintf(vrpnConnectionAddress, "%s:%s", szServerAddress, szPortNumber);
	vrpn_Connection* connection = vrpn_get_connection_by_name(vrpnConnectionAddress);

	// fire it up
	int retCode = 0;

		// Create a tracker (aka RigidBody)
	sprintf(szObjectName, "%s@%s:%s", szObjectName, szServerAddress, szPortNumber);
	printf("Connecting to Motive RigidBody object %s...\n", szObjectName);
	vrpn_Tracker_Remote* vrpnTracker = new vrpn_Tracker_Remote(szObjectName);
	if (vrpnTracker)
	{
	vrpnTracker->set_update_rate(kTrackerSampleRate);
	retCode = vrpnTracker->register_change_handler(0, handle_tracker);
		if (retCode != 0)
			printf(" => Connect failed :  %d\n", retCode);
		else
			printf(" => Connection succeeded.\n");
	}
	else
	{
		printf("Create Tracker Error - skipping.\n");
	}

	// Create an 'Axis' (aka 'analog')
	sprintf(szObjectName, "%s@%s", kAnalogName, szServerAddress);
	printf("Connecting to Motive Peripheral Analog Device %s...\n", szObjectName);
	vrpn_Analog_Remote* vrpnAnalog = new vrpn_Analog_Remote(szObjectName);
	if (vrpnAnalog)
	{
		vrpnAnalog->shutup = true;
		retCode = vrpnAnalog->register_change_handler(0, handle_analog);
		if (retCode != 0)
			printf(" => Connect failed :  %d\n", retCode);
		else
			printf(" => Connection succeeded.\n");
	}
	else
	{
		printf("Create Analog Error - skipping.\n");
	}

	// Create a button
	sprintf(szObjectName, "%s@%s", kButtonName, szServerAddress);
	printf("Connecting to Motive Peripheral Button Device %s...\n", szObjectName);
	vrpn_Button_Remote* vrpnButton = new vrpn_Button_Remote(szObjectName);
	if (vrpnButton)
	{
		vrpnButton->shutup = true;
		retCode = vrpnButton->register_change_handler(0, handle_button);
		if (retCode != 0)
			printf(" => Connect failed :  %d\n", retCode);
		else
			printf(" => Connection succeeded.\n");
	}
	else
	{
		printf("Create Button Error - skipping.\n");
	}

	// Capture data to output files for 10.0 seconds
	printf("\nCollecting %.2f seconds of data...\n", kCaptureDuration);
	HiTimer tm;
	tm.Start();
	while (tm.Elapsed() < kCaptureDuration)
	{
		if (vrpnTracker)
			vrpnTracker->mainloop();
		if (vrpnAnalog)
			vrpnAnalog->mainloop();
		if (vrpnButton)
			vrpnButton->mainloop();

		vrpn_SleepMsecs(0);
	}

	// cleanup
	if (vrpnTracker)
	{
		vrpnTracker->unregister_change_handler(0, handle_tracker);
		delete vrpnTracker;
	}
	if (vrpnAnalog)
	{
		vrpnAnalog->unregister_change_handler(0, handle_analog);
		delete vrpnAnalog;
	}
	if (vrpnButton)
	{
		vrpnButton->unregister_change_handler(0, handle_button);
		delete vrpnButton;
	}		


	// Close up timing files
	if (gTrackerFile)
	{
		fprintf(gTrackerFile, "\nTracker Dropped Frames: %d,Duplicate Frames: %d, Received Frames: %d", gTrackerDroppedFrames, gTrackerDuplicateFrames, gTrackerReceivedFrames);
		fflush(gTrackerFile);
		fclose(gTrackerFile);
	}

	if (gAnalogFile)
	{
		fprintf(gAnalogFile, "\nAnalog Dropped Frames: %d,Duplicate Frames: %d, Received Frames: %d", gAnalogDroppedFrames, gAnalogDuplicateFrames, gAnalogReceivedFrames);
		fflush(gAnalogFile);
		fclose(gAnalogFile);
	}

	if (gButtonFile)
	{
		fprintf(gButtonFile, "\nButton Dropped Frames: %d,Duplicate Frames: %d, Received Frames: %d", gButtonDroppedFrames, gButtonDuplicateFrames, gButtonReceivedFrames);
		fflush(gButtonFile);
		fclose(gButtonFile);
	}


	//obtain the percentage 

	printf("\nCollecting Complete. \nTiming data files written to %s\n", szFolder);
	printf("Tracker Dropped Frames: %d\n", gTrackerDroppedFrames);
	printf("Duplicate Frames: %d\n", gTrackerDuplicateFrames);
	printf("Recieved Frames: %d\n", gTrackerReceivedFrames);
	printf("\nPress any key to quit.");
	while (!_kbhit())
	{
		std::this_thread::sleep_for(std::chrono::milliseconds(0));
	}

	return 0;
}
